{-----------------------------------------------------------------------}
{ Ansi.pas is Copyright (c) 1992 by L. B. Neal. All Rights Reserved.    }
{ Portions of this code are based on Public Domain code by Jon Ebert.   }
{ Permission is granted for personal non-commercial use by individuals. }
{ NOTICE: This code is presented AS-IS without any form of warranty.    }
{         Use of this code confirms agreement with this notice.         }
{-----------------------------------------------------------------------}
{ NOTE: This UNIT was NOT designed to be stand-alone and thus will      }
{       present a variety of problems when interfaced with other        }
{       programs.  I would hope the concepts and optimizations would    }
{       lead the user to ideas and implementations of even better       }
{       and faster versions of ansi.pas.                                }
{-----------------------------------------------------------------------}
{ Comments are minimal in this section to make the user/programmer think}

UNIT ANSI;
{$V-,D+,L+,S-,R-,B-}{2.23}

Interface

PROCEDURE PUT_ANSI;
PROCEDURE Windo(StartX,StartY,EndX,EndY: Byte);

Implementation
USES DOS,CRT,COMMON;

CONST
     Max_Cursor_Stack = 100;
     Cur_Cursor_Stack : Byte = 1;
     num              : Byte = 1;
     Color: ARRAY[1..8] OF Byte = (Black,Red,Green,Brown,Blue,Magenta,Cyan,LightGray);
     Up    = 1;
     Down  = 2; 
     Right = 3;
     Left  = 4;
     Place = 5;
     Dir:  ARRAY[1..5] OF Byte = (Up,Down,Right,Left,Place);
     ANSIStrt : String[4] = #27+'[';   {2.24}

VAR
    nstack: ARRAY[1..20] OF ShortInt;
    x_cursor_stack: ARRAY[1..Max_Cursor_Stack] OF Byte;
    y_cursor_stack: ARRAY[1..Max_Cursor_Stack] OF Byte;

    RFstCol,RFstRow,RLstCol,RLstRow: Byte;
    FstCol,FsTRow,LstCol,LstRow: Byte;

 PROCEDURE Windo(StartX, StartY, EndX, EndY : Byte);
 BEGIN
  FstCol := StartX; LstCol := EndX; {0.95}
  FstRow := StartY; LstRow := EndY; {0.95}
  RFstCol := StartX-1; RLstCol := EndX-1;
  RFstRow := StartY-1; RLstRow := EndY-1;
  Window(StartX, StartY, EndX, EndY);
 END;

 PROCEDURE BGotoXy(gcol,grow: Byte); { GoToXY in this window }
 VAR wrow,wcol: Byte;
 BEGIN
  DEC(gcol); DEC(grow);
  wcol := RFstCol+gcol; wrow := RFstRow+grow;
  INLINE(                         { GotoXY BIOS   }
   $8A/$B6/wrow/                  { MOV DH,wrow   }
   $8A/$96/wcol/                  { MOV DL,wcol   }
   $B4/$02/                       { MOV AH,02     }
   $30/$FF/                       { XOR BH,BH     }
   $55/                           { PUSH BP       }
   $CD/$10/                       { INT 10H       }
   $5D);                          { POP BP        }
 END;

 PROCEDURE Cursor_Move(way:byte);
 VAR n1,n2,thiscol,thisrow: Byte;
 BEGIN
  IF nstack[1] > 0 THEN n1 := nstack[1] ELSE n1 := 1;
  thiscol := WhereX; thisrow := WhereY;
  CASE way OF
   Left  : IF (thiscol-n1) >= FstCol THEN  {4}
            BGotoXY(thiscol-n1,thisrow)
           ELSE
            BGotoXY(FstCol,thisrow);
   Place : BEGIN {5}
            IF nstack[2] > 0 THEN n2 := nstack[2] ELSE n2 := 1;
            BGotoXY(n2,n1);
           END;
   Up    : IF (thisrow-n1) >= FstRow THEN  {1}
            BGotoXY(thiscol,thisrow-n1)
           ELSE
            BGotoXY(thiscol,FstRow);
   Down  : IF (thisrow+n1) <= LstRow THEN  {2}
            BGotoXY(thiscol,thisrow+n1)
           ELSE
            BGotoXY(thiscol,LstRow);
   Right : IF (thiscol+n1) <= LstCol THEN  {3}
            BGotoXY(thiscol+n1,thisrow)
           ELSE
            BGotoXY(LstCol,thisrow);
  END;{case}
 END;{Cursor_Move}

 FUNCTION Push_Cursor: Boolean;
 BEGIN
  IF (num > 1) OR (nstack[1] >= 0) THEN
   Push_Cursor := False
  ELSE 
   BEGIN
    x_cursor_stack[Cur_Cursor_Stack] := WhereX;
    y_cursor_stack[Cur_Cursor_Stack] := WhereY;
    IF (Cur_Cursor_Stack < MAX_CURSOR_STACK) THEN INC(Cur_Cursor_Stack);
    Push_Cursor := True;
   END;
 END;

 FUNCTION Pop_Cursor: Boolean;
 BEGIN
  IF (num > 1) OR (nstack[1] >= 0) THEN
   Pop_Cursor := False
  ELSE
   BEGIN
    DEC(cur_cursor_stack);
    IF cur_cursor_stack < 1 THEN Cur_Cursor_Stack := 1;
    GotoXY(x_cursor_stack[cur_cursor_stack],
          y_cursor_stack[cur_cursor_stack]);
    Pop_Cursor := True;
   END;
 END;

 FUNCTION Clear_Screen: Boolean;
 BEGIN
  IF (num > 1) OR (nstack[1] <> 2) THEN
   Clear_Screen := False
  ELSE
   BEGIN
    ClrScr;
    Clear_Screen := True;
   END;
 END;

 FUNCTION Clear_To_Eol: Boolean;
 BEGIN
  IF (nstack[1] >= 0) THEN
   Clear_To_Eol := False
  ELSE
   BEGIN
    ClrEol;
    Clear_To_Eol := True;
   END;
 END;

 PROCEDURE Display_Attr;{set display attributes}
 VAR oldintense,oldblink,snum,index,cnt: Byte;
 BEGIN
  index := num;
  cnt := 1;
  REPEAT
   snum := nstack[cnt];
   IF ( (snum > 29) AND (snum < 38) ) THEN
    TextColor( (TextAttr AND 8) OR (Color[snum-29]) )
   ELSE  
    IF ( (snum < 48) AND (snum > 39) ) THEN
     TextBackGround( ((TextAttr SHR 4) AND 8) OR (Color[snum-39]) )
    ELSE
     CASE snum OF
      0 : TextAttr := 7;
      1 : TextAttr := TextAttr OR 8;            {Intensified}
      4 : IF scrmode = 7 THEN TextAttr := 1;    {Underline MONO only}
      5 : TextAttr := TextAttr OR 128;          {Blink}
      7 : BEGIN TextColor(LightGray); TextBackGround(Black); END;
      8 : TextColor( (TextAttr SHR 4) AND 128); {invisable}
    END;{case}
   INC(cnt);
  UNTIL (cnt > index) OR (snum < 0);
 END;

 FUNCTION Display_Mode: Boolean; {DON'T DO IT!}
 BEGIN
  Display_Mode := True; {default}
  IF (num < 2) THEN Display_Mode := False;
  IF (nstack[2] <> 0) THEN Display_Mode := False;
 END;

 FUNCTION truncate_80: Boolean; {truncate at col80 NOP}
 BEGIN
  IF (num <> 2) OR (nstack[2] <> 7) OR (nstack[1] <> 0) THEN
   Truncate_80 := False
  ELSE
   Truncate_80 := True;   {Don't do anything}
 END;

 FUNCTION Redefine_Key: Boolean; {Don't do it}
 BEGIN
  Redefine_Key := True;
 END;

 FUNCTION AnsiRep: Boolean; {2.24 added}
 VAR Reply: String[20]; Col,Row: String[5];
 BEGIN
  STR(WhereX,col);
  STR(WhereY,row);
  Reply := #27+'['+col+';'+row+'R';
  SendStr(Reply);
  AnsiRep := True;
 END;

{NOTE: This routine MUST handle incoming characters from the com port.}
{      As presented it interfaces with TMODEM.                        }

{$F+}
 FUNCTION Ainkey: Char;{$F-} {2.23}
 BEGIN
  inline($FA);
  Ainkey := RecBuf[buffer_Tail];
  IF Buffer_Tail < Buffer_End THEN INC(Buffer_Tail) ELSE Buffer_Tail := 1;
  DEC(buffer_count);
  IF Buffer_Count = 0 THEN Inready := False;
  inline($FB);
  port[$20] := $20;
 END;

 { -------- This is the guts of ANSI.PAS -------------- }
 PROCEDURE PUT_ANSI;
 VAR done: Boolean; cnt,oac: Byte; ac: Char;
 BEGIN
  done := False;
  num := 20;
  WHILE num > 0 DO BEGIN nstack[num] := -1; DEC(num); END;
  Num := 1;
  cnt := 1;
  IF inready THEN
   ac := Ainkey
  ELSE
   BEGIN
    WHILE NOT InReady AND (cnt < 101) DO
     BEGIN
      Delay(1); INC(cnt);
     END;
    IF Inready THEN ac := Ainkey;
   END;
  IF ac = '[' THEN
   BEGIN
    { ------ parsing loop -------- }
    REPEAT{until done}
     IF Inready THEN
      ac := Ainkey
     ELSE
      BEGIN
       cnt := 1;
       WHILE NOT InReady AND (cnt < 101) DO
        BEGIN
         Delay(1); INC(cnt);
        END;
       IF Inready THEN ac := Ainkey;
      END;
     oac := ORD(ac);
     IF ( (oac >= 48) AND (oac <= 57) ) THEN
      BEGIN
       IF nstack[num] >= 0 THEN
        nstack[num] := (nstack[num]*10)+(oac-48)
       ELSE
        nstack[num] := oac-48; {ok here}
      END
     ELSE
      CASE oac OF
        59 : INC(num); {;} {ok skip to next num}
       109 : BEGIN Display_Attr; Done := True; END; {m}
        75 : Done := Clear_To_Eol;      {K}
        74 : Done := Clear_Screen;      {J}
    102,72 : BEGIN     {f,H}
              Cursor_Move(dir[place]); Done := True;
             END;
    65..68 : BEGIN     {A..D}
              Cursor_Move(dir[ORD(ac)-64]); Done := True;
             END;
       115 : Done := Push_Cursor;       {s}
       117 : Done := Pop_Cursor;        {u}
       104 : Done := Display_Mode;      {h}
       108 : Done := Truncate_80;       {l}
       112 : Done := Redefine_Key;      {p}
        61 : IF (num = 1) THEN BEGIN INC(num); nstack[num] := 0; END;
       110 : Done := AnsiRep;           {n} {2.24 added}
      END;{case oac}
     UNTIL ((port[cdetport] AND 128) = 0) OR Done; {2.23}
    
    { ------------- end parsing loop ---------------- }
   END;
 END;{PUT_ANSI}

END.
